
#ifndef __IPSEC_H__
#define __IPSEC_H__



typedef unsigned char 	cookie_t[8];


/*
 * internet security association and key management protocal
 * message header...
 */
struct isakmp_hdr {
    cookie_t	init_cookie;
    cookie_t	resp_cookie;
    __u8		next_payload;
    __u8		version;
    __u8		exchange_type;
    __u8		encryption:1;
    __u8		commit:1;
    __u8		auth:1;
    __u8		other_bits:5;
    __u32		msg_id;
    __u32		length;
} __attribute__ ((packed));

#define SA_PAYLOAD	1
#define EXCHANGE_TYPE_ID_PROTECT	2
#define DOI_IPSEC	1UL
#define PROTO_ISAKMP	1


#define HMAC_MD5_BLOCK_LEN		64
#define HMAC_SHA1_BLOCK_LEN		64
#define HMAC_MD5_OUTPUT_LEN		16
#define HMAC_SHA1_OUTPUT_LEN	20	
#define DES3_CBC_BLOCK_LEN		8


struct sa_payload {
    __u8	next_payload;
    __u8	reserved;
    __u16	length;
    __u32	doi;
    __u32	situation;
} __attribute__ ((packed));


struct proposal_payload {
    __u8	next_payload;
    __u8	reserved;
    __u16	length;
    __u8	proposal_number;
    __u8	proto_id;
    __u8	spi_size;
    __u8	transform_count;
} __attribute__ ((packed));


struct transform_payload {
    __u8	next_payload;
    __u8	reserved;
    __u16	length;
    __u8	transform_number;
    __u8	transform_id;
    __u16	zero;
    __u32	life_type;
    __u32	life_duration;
    __u32	crypto_algo;
    __u32	hash_algo;
    __u32	auth_method;
    __u32	group_desc;
} __attribute__ ((packed));

struct transform_payload2 {
    __u8	next_payload;
    __u8	reserved;
    __u16	length;
    __u8	transform_number;
    __u8	transform_id;
    __u16	zero;
    __u16	group_desc;
    __u16	group_num;
    __u16	encap_desc;
    __u16	encap_mode;
    __u16	life_desc;
    __u16	life_secs;
    __u16	duration_desc;
    __u16	duration_secs;
    __u16	auth_desc;
    __u16	auth_mode;
} __attribute__ ((packed));

struct key_exchange_payload {
    __u8	next_payload;
    __u8	reserved;
    __u16	length;
} __attribute__ ((packed));

struct nonce_payload {
    __u8	next_payload;
    __u8	reserved;
    __u16	length;
} __attribute__ ((packed));

struct isakmp_nonce {
    __u8	next_payload;
    __u8	reserved;
    __u16	length;
    __u8	nonce[16];
} __attribute__ ((packed));

struct isakmp_ke {
    __u8	next_payload;
    __u8	reserved;
    __u16	length;
    __u8	ke[192];
} __attribute__ ((packed));

struct isakmp_1 {
    struct isakmp_hdr			hdr;
    struct sa_payload			sa;
    struct proposal_payload		proposal;
    struct transform_payload	transform;
} __attribute__ ((packed));

struct isakmp_2 {
    struct isakmp_hdr			hdr;
    struct key_exchange_payload	key_exchange;
    __u8						key_data[192];
    struct nonce_payload		nonce;
    __u8						nonce_data[16];
} __attribute__ ((packed));


struct isakmp_id {
    __u8	next_payload;
    __u8	reserved;
    __u16	length;
    __u8	id_type;
    __u8	proto_id;
    __u16	port;
    __u32	ipaddr;
} __attribute__ ((packed));

struct isakmp_id_subnet {
    __u8	next_payload;
    __u8	reserved;
    __u16	length;
    __u8	id_type;
    __u8	proto_id;
    __u16	port;
    __u32	ipaddr;
    __u32	netmask;
} __attribute__ ((packed));

struct isakmp_hash {
    __u8	next_payload;
    __u8	reserved;
    __u16	length;
    __u8	hash[HMAC_MD5_OUTPUT_LEN];
} __attribute__ ((packed));

struct isakmp_sig {
    __u8	next_payload;
    __u8	reserved;
    __u16	length;
    __u8	sig[RSA_MODULUS_LEN];
} __attribute__ ((packed));

struct isakmp_3 {
    struct isakmp_hdr	hdr;
    struct isakmp_id	id;
    struct isakmp_sig	sig;
} __attribute__ ((packed));


struct isakmp_4 {
    struct isakmp_hdr			hdr;
    struct isakmp_hash			hash;
    struct sa_payload			sa;
    struct proposal_payload		proposal;
    __u32						spi;
    struct transform_payload2	transform;
    struct isakmp_nonce			nonce;
    struct isakmp_ke			ke;
    struct isakmp_id_subnet		id1;
    struct isakmp_id_subnet		id2;
} __attribute__ ((packed));

struct isakmp_5 {
    struct isakmp_hdr	hdr;
    struct isakmp_hash	hash;
} __attribute__ ((packed));


typedef __u8	skeyid_t[16];
typedef __u8	des3_key_t[24];


struct hash_i {
    __u8		g_i[192];
    __u8		g_r[192];
    cookie_t	init_cookie;
    cookie_t	resp_cookie;
    __u32		doi;
    __u32		situation;
    __u8		next_payload;
    __u8		reserved;
    __u16		length;
    __u8		proposal_number;
    __u8		proto_id;
    __u8		spi_size;
    __u8		transform_count;
    __u8		next_payload_;
    __u8		reserved_;
    __u16		length_;
    __u8		transform_number;
    __u8		transform_id;
    __u16		zero;
    __u32		life_type;
    __u32		life_duration;
    __u32		crypto_algo;
    __u32		hash_algo;
    __u32		auth_method;
    __u32		group_desc;
    __u8		id_type;
    __u8		proto_id_;
    __u16		port;
    __u32		ipaddr;
} __attribute__ ((packed));


struct quick_sa_context {
    DH					*dh;
    __u8				n_i[16];
    __u8				n_r[16];
    __u32				msg_id;
    __u8				our_iv[HMAC_MD5_OUTPUT_LEN];
    __u8				their_iv[HMAC_MD5_OUTPUT_LEN];
    __u8				our_keymat[40];
    __u8				their_keymat[40];
    //des_cblock			our_keymat[5];
    //des_cblock			their_keymat[5];
    des_key_schedule	our_k1;
    des_key_schedule	our_k2;
    des_key_schedule	our_k3;
    des_key_schedule	their_k1;
    des_key_schedule	their_k2;
    des_key_schedule	their_k3;
    __u8				iv[HMAC_MD5_OUTPUT_LEN];
    __u32				our_spi;
    __u32				their_spi;
    __u8				dh_secret[192];
};

struct ipsec_sa_context {
    cookie_t			init_cookie;
    cookie_t			resp_cookie;
    int					state;
    unsigned int		seq;
    DH					*dh;
    __u8				n_i[16];
    __u8				n_r[16];
    __u8				dh_priv[192];
    __u8				dh_pub[192];
    __u8				dh_other_pub[192];
    __u8				dh_secret[192];

    /* keys */
    skeyid_t			skeyid;
    skeyid_t			skeyid_d;
    skeyid_t			skeyid_a;
    skeyid_t			skeyid_e;
    des_cblock			enc_key[3];
    des_key_schedule	k1;
    des_key_schedule	k2;
    des_key_schedule	k3;
    __u8				des3_iv[HMAC_MD5_OUTPUT_LEN];
    __u8				other_iv[HMAC_MD5_OUTPUT_LEN];
    __u8				hash_i[HMAC_MD5_OUTPUT_LEN];
    __u8				sig_i[RSA_MODULUS_LEN];

    struct quick_sa_context	qk_ctx;
};

struct esp_hdr
{
    __u32   esp_spi;        
    __u32   esp_rpl;       
    __u8    esp_iv[8];    
} __attribute__ ((packed));


#define ISAKMP_NEXT_NONE       0	/* No other payload following */
#define ISAKMP_NEXT_SA         1	/* Security Association */
#define ISAKMP_NEXT_P          2	/* Proposal */
#define ISAKMP_NEXT_T          3	/* Transform */
#define ISAKMP_NEXT_KE         4	/* Key Exchange */
#define ISAKMP_NEXT_ID         5	/* Identification */
#define ISAKMP_NEXT_CERT       6	/* Certificate */
#define ISAKMP_NEXT_CR         7	/* Certificate Request */
#define ISAKMP_NEXT_HASH       8	/* Hash */
#define ISAKMP_NEXT_SIG        9	/* Signature */
#define ISAKMP_NEXT_NONCE      10	/* Nonce */
#define ISAKMP_NEXT_N          11	/* Notification */
#define ISAKMP_NEXT_D          12	/* Delete */
#define ISAKMP_NEXT_VID        13	/* Vendor ID */


/* group 5 prime modulus */
#define MODP1536_MODULUS \
    "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xC9\x0F\xDA\xA2\x21\x68\xC2\x34\xC4\xC6\x62\x8B\x80\xDC\x1C\xD1" \
    "\x29\x02\x4E\x08\x8A\x67\xCC\x74\x02\x0B\xBE\xA6\x3B\x13\x9B\x22\x51\x4A\x08\x79\x8E\x34\x04\xDD" \
    "\xEF\x95\x19\xB3\xCD\x3A\x43\x1B\x30\x2B\x0A\x6D\xF2\x5F\x14\x37\x4F\xE1\x35\x6D\x6D\x51\xC2\x45" \
    "\xE4\x85\xB5\x76\x62\x5E\x7E\xC6\xF4\x4C\x42\xE9\xA6\x37\xED\x6B\x0B\xFF\x5C\xB6\xF4\x06\xB7\xED" \
    "\xEE\x38\x6B\xFB\x5A\x89\x9F\xA5\xAE\x9F\x24\x11\x7C\x4B\x1F\xE6\x49\x28\x66\x51\xEC\xE4\x5B\x3D" \
    "\xC2\x00\x7C\xB8\xA1\x63\xBF\x05\x98\xDA\x48\x36\x1C\x55\xD3\x9A\x69\x16\x3F\xA8\xFD\x24\xCF\x5F" \
    "\x83\x65\x5D\x23\xDC\xA3\xAD\x96\x1C\x62\xF3\x56\x20\x85\x52\xBB\x9E\xD5\x29\x07\x70\x96\x96\x6D" \
    "\x67\x0C\x35\x4E\x4A\xBC\x98\x04\xF1\x74\x6C\x08\xCA\x23\x73\x27\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF"

#define MODP_GENERATOR "\x02"


#endif /* #ifndef __IPSEC_H__ */

