#include <sys/types.h>
#include <asm/types.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <features.h>    /* for the glibc version number */
#include <netpacket/packet.h>
#include <netinet/in.h>
#include <net/ethernet.h>     /* the L2 protocols */
#include <syslog.h>
#include <ctype.h>
#include <limits.h>
#ifndef __USE_BSD
#define __USE_BSD
#endif
#ifndef __FAVOR_BSD
#define __FAVOR_BSD
#endif
#include <arpa/inet.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>

#include "80211.h"
#include "../airjack.h"

#pragma pack(1)



/*** The Marshmello People ***/

struct ip_80211_hdr {
    struct a4_80211		llh;
    struct llc_hdr		llc;
    struct ip			iph;
} __attribute__ ((packed));

struct ip_8023_hdr {
    struct ether_header	llh;
    struct ip			iph;
} __attribute__ ((packed));

struct udp_80211_hdr {
    struct a4_80211		llh;
    struct llc_hdr		llc;
    struct ip			iph;
    struct udphdr		udph;
} __attribute__ ((packed));

struct udp_8023_hdr {
    struct ether_header	llh;
    struct ip			iph;
    struct udphdr		udph;
} __attribute__ ((packed));


struct pseudo_hdr {
    __u32	src_ip;
    __u32	dst_ip;
    __u8	zero;
    __u8	proto;
    __u16	len;
} __attribute__ ((packed));


/*
 * as usual this is shamelessly stolen from ping.c
 * this assumes a packet less than 32k...
 */
static __inline__ __u16 in_cksum (__u16 *p, unsigned int len)
{
    register unsigned int	sum = 0;
    register int			nwords = len >> 1;

    while (nwords-- != 0)
        sum += *p++;

    if (len & 1) {
        union {
            u_short w;
            u_char c[2];
        } u;
        u.c[0] = *(u_char *)p;
        u.c[1] = 0;
        sum += u.w;
    }

    /* end-around-carry */
    sum = (sum >> 16) + (sum & 0xffff);
    sum += (sum >> 16);
    return (~sum);
}


/*
 * appends an ip header to the buffer, adds the correct checksum and all that...
 */
size_t ip_encap (void *buf, size_t payload_len, __u32 saddr, __u32 daddr, __u8 proto)
{
    struct ip	*iph = (struct ip *)buf;
    

    /* hope you reserved enough space */
    memmove(buf + sizeof(struct ip), buf, payload_len);
    memset(buf, 0, sizeof(struct ip));

    iph->ip_hl			= 5;			/* no options */
    iph->ip_v			= 4;			/* ipv4 */
    iph->ip_tos			= 0;			/* bah */
    iph->ip_len			= htons(payload_len + sizeof(struct ip));
    iph->ip_id			= rand();		/* tims mother */
    iph->ip_off			= 0;			/* again, tims mother */
    iph->ip_ttl			= 30;			/* still tims mother */
    iph->ip_p			= proto;
    iph->ip_sum			= 0;
    iph->ip_src.s_addr	= saddr;  
    iph->ip_dst.s_addr	= daddr;	/* broadcast to tims mother */

    /* ip check sum */
    iph->ip_sum			= in_cksum((__u16 *)iph, sizeof(struct ip));

    return(payload_len + sizeof(struct ip));
}


/*
 * appends an 802.3 header to the buffer...
 */
size_t ll8023_encap (void *buf, size_t payload_len, __u8 *saddr, __u8 *daddr, __u16 type)
{
    struct ether_header	*llh = (struct ether_header *)buf;

    
    /* hope you reserved enough space */
    memmove(buf + sizeof(struct ether_header), buf, payload_len);
    memset(buf, 0, sizeof(struct ether_header));

    memcpy(llh->ether_shost, saddr, 6);
    memcpy(llh->ether_dhost, daddr, 6);
    llh->ether_type = htons(type);
    return(payload_len + sizeof(struct ether_header));
}


/*
 * encapsulates a packet in an 802.11 header...
 */
size_t ll80211_encap (void *buf, size_t payload_len, __u8 *saddr, __u8 *daddr, __u8 *bssid, __u16 type, __u8 from_ds, __u8 to_ds)
{
    struct a3_80211	*llh = (struct a3_80211 *)buf;
    struct llc_hdr	*llc = (struct llc_hdr *)(llh+1);


    /* hope you reserved enough space */
    memmove(buf + sizeof(struct a3_80211) + sizeof(struct llc_hdr), buf, payload_len);
    memset(buf, 0, sizeof(struct a3_80211) + sizeof(struct llc_hdr));

    memcpy(llh->mh_mac1, daddr, 6);
    memcpy(llh->mh_mac2, bssid, 6);
    memcpy(llh->mh_mac3, saddr, 6);

    /* basicly a magic number */
    memcpy(llc, "\xaa\xaa\x03\x00\x00\x00", 6);

    llc->type = htons(type);

    llh->mh_type = FC_TYPE_DATA;
    llh->mh_subtype = DATA_DATA;
    llh->mh_from_ds = from_ds;
    llh->mh_to_ds = to_ds;
    return(payload_len + sizeof(struct a3_80211) + sizeof(struct llc_hdr));
}


/*
 * encapsulates the packet as a udp datagram, sets up proper checksums and all that...
 *
 * NOTE: 
 *    im lazy, so im just going to use the buffer as a scratch pad to do the pseudo header
 *    for the checksum, so you better have a bit of extra space in the buffer...
 */
size_t udp_encap (void *buf, size_t payload_len, __u32 saddr, __u32 daddr, __u16 sport, __u16 dport)
{
    struct pseudo_hdr	*ph = (struct pseudo_hdr *)buf;
    struct udphdr		*udph = (struct udphdr *)(ph+1);

    /* hope you reserved enough space */
    memmove(((__u8 *)buf) + sizeof(struct pseudo_hdr) + sizeof(struct udphdr), buf, payload_len);
    memset(buf, 0, sizeof(struct pseudo_hdr) + sizeof(struct udphdr));

    udph->uh_sport		= htons(sport);	/* tims mother uses this one for source */
    udph->uh_dport		= htons(dport);	/* and this one for destination */
    udph->uh_ulen		= htons(sizeof(struct udphdr) + payload_len);
    udph->uh_sum		= 0;			

    ph->src_ip			= saddr;
    ph->dst_ip			= daddr;
    ph->proto			= IPPROTO_UDP;
    ph->len				= htons(sizeof(struct udphdr) + payload_len);
    udph->uh_sum		= in_cksum((u_short *)ph, sizeof(struct pseudo_hdr) + payload_len + sizeof(struct udphdr));

    /* ok, now we're finished with the pseudo header */
    memmove(buf, ((__u8 *)buf) + sizeof(struct pseudo_hdr), payload_len + sizeof(struct udphdr));
    memset(buf + payload_len + sizeof(struct udphdr), 0, sizeof(struct pseudo_hdr));

    return(payload_len + sizeof(struct udphdr));
}


/*
 * returns a pointer to the start of the payload for the given ip packet...
 *
 * blah, crc error checking, you're on your own...
 */
void *ip_8023_payload (void *buf, size_t buflen)
{
    struct ip_8023_hdr	*packet = buf;


    if(buflen < sizeof(*packet)) {
        return(NULL);
    }

    /* ip options */
    if(buflen < sizeof(struct ether_header) + (packet->iph.ip_hl*4)) {
        return(NULL);
    }

    if(packet->llh.ether_type != htons(ETHERTYPE_IP)) {
        return(NULL);
    }

    return(buf + sizeof(struct ether_header) + (packet->iph.ip_hl*4));
}


/*
 * returns a pointer to the start of the payload for the given udp datagram...
 *
 * blah, crc error checking, you're on your own...
 */
void *udp_8023_payload (void *buf, size_t buflen)
{
    struct udp_8023_hdr	*packet = buf;


    if(buflen < sizeof(*packet)) {
        return(NULL);
    }

    /* ip options */
    if(buflen < sizeof(struct ether_header) + sizeof(struct udphdr) + (packet->iph.ip_hl*4)) {
        return(NULL);
    }

    if(packet->llh.ether_type != htons(ETHERTYPE_IP)) {
        return(NULL);
    }

    if(packet->iph.ip_p != IPPROTO_UDP) {
        return(NULL);
    }
    
    return(buf + sizeof(struct ether_header) + sizeof(struct udphdr) + (packet->iph.ip_hl*4));
}


/*
 * returns a pointer to the start of the payload for the given udp datagram...
 *
 * blah, crc error checking, you're on your own...
 */
void *ip_80211_payload (void *buf, size_t buflen)
{
    struct ip_80211_hdr	*packet = buf;


    if(buflen < sizeof(*packet)) {
        return(NULL);
    }

    /* ip options */
    if(buflen < sizeof(struct a4_80211) + sizeof(struct llc_hdr) + (packet->iph.ip_hl*4)) {
        return(NULL);
    }

    if(packet->llc.type != htons(ETHERTYPE_IP)) {
        return(NULL);
    }

    return(buf + sizeof(struct a4_80211) + sizeof(struct llc_hdr) + (packet->iph.ip_hl*4));
}


/*
 * returns a pointer to the start of the payload for the given udp datagram...
 *
 * blah, crc error checking, you're on your own...
 */
void *udp_80211_payload (void *buf, size_t buflen)
{
    struct udp_80211_hdr	*packet = buf;


    if(buflen < sizeof(struct udp_80211_hdr)) {
        return(NULL);
    }

    /* ip options */
    if(buflen < sizeof(struct a4_80211) + sizeof(struct llc_hdr) + sizeof(struct udphdr) + (packet->iph.ip_hl*4)) {
        return(NULL);
    }

    if(packet->llc.type != htons(ETHERTYPE_IP)) {
        return(NULL);
    }

    if(packet->iph.ip_p != IPPROTO_UDP) {
        return(NULL);
    }
    
    return(buf + sizeof(struct a4_80211) + sizeof(struct llc_hdr) + sizeof(struct udphdr) + (packet->iph.ip_hl*4));
}






